﻿------------------------------------------------------------------------------
-- $Id$
------------------------------------------------------------------------------
--                           clcInfo_Warlock.lua
--
--	Copyright (C) 2012 Deronda, Maighurl, Doogal
--
--  This file is for use with clcInfo by abija.
--
--  abija has distributed clcInfo All Rights Reserved.  For more information
--  go to: http://wow.curseforge.com/addons/clctracker/
--
------------------------------------------------------------------------------
-- Author:
--      Deronda (@ Stormreaver), Maighurl, Doogal
------------------------------------------------------------------------------
-- Description:
--   Global namespace for clcInfo_Warlock.  The purpose is to avoid repetition.
--   However, since this module generates icons and bars that update many times
--   per second, locals should be used whenever possible and table lookups
--   avoided to speed things up.
------------------------------------------------------------------------------
if not clcInfo then return end       -- need clcInfo to run
if clcInfo_Warlock then return end   -- don't load if already loaded
local ENGLISH_CLASS_NAME = "WARLOCK"
local _, englishClass  = UnitClass("player") -- only load for locks
if englishClass ~= ENGLISH_CLASS_NAME then return end

-----------------NAMESPACE---------------------------------------
clcInfo_Warlock = {}
clcInfo_Warlock.VERSION = 0.1 -- bump this up when fields are added to the tdb
                              -- or db, or priorities are changed in some way.
clcInfo_Warlock.CLASS = ENGLISH_CLASS_NAME

-----------------LIBRARIES---------------------------------------
local range_check = LibStub:GetLibrary("LibRangeCheck-2.0", true)
if not range_check then
	print("clcInfo_Warlock ERROR: LibRangeCheck-2.0 failed to load.")
    clcInfo_Warlock = nil
    return
end

----------------WARLOCK SPELLS-----------------------------------
--Though dot clips used to be a user setting, that has been removed
--and replaced by this.  The addon should be responsible for the
--optimal clip time.
clcInfo_Warlock.CLIP_HARMFUL = 2 -- clip harmful ticks at 2 seconds

--put any spell you need to reference in templates or modules here.
clcInfo_Warlock.spells = {
    --Affliction Only
    ["Unstable Affliction"] = 30108,
    ["Haunt"]               = 48181,
    --Demonology Only
    ["Felstorm"]            = 89751,
    ["Decimation"]          = 63167,
    ["Demon Leap"]          = 54785,
    ["Demon Soul:Felguard"]  = 79462,
    ["Demonic Empowerment"] = 47193,
    ["Hand of Guldan"]      = 71521,
    ["Immolation Aura"]     = 50589,
    ["Metamorphosis"]       = 59672,
    ["Molten Core"]         = 71165,
    ["Summon Felguard"]     = 30146,
    --Destruction Only
    ["Backdraft"]           = 54277,
    ["Chaos Bolt"]          = 50796,
    ["Conflagrate"]         = 17962,
    ["Empowered Imp"]       = 47220,
    ["Improved Soul Fire"]  = 18119,
    ["Shadowburn"]          = 17877,
    ["Shadowfury"]          = 30283,
    --generic lock
    ["Bane of Agony"]       = 980,
    ["Bane of Doom"]        = 603,
    ["Corruption"]          = 172,
    ["Dark Intent"]         = 85768,
    ["Demon Soul"]          = 77801,
    ["Demon Soul:Felhunter"] = 79460,
    ["Demon Soul:Imp"]       = 79459,
    ["Demon Soul:Succubus"]  = 79463,
    ["Demon Soul:Voidwalker"]= 79464,
    ["Drain Soul"]          = 1120,
    ["Fel Armor"]           = 28176,
    ["Immolate"]            = 348,
    ["Incinerate"]          = 29722,
    ["Life Tap"]            = 1454,
    ["Shadow Bolt"]         = 686,
    ["Shadowflame"]         = 47897,
    ["Summon Doomguard"]    = 18540,
    ["Summon Felhunter"]    = 691,
    ["Summon Imp"]          = 688,
    ["Summon Infernal"]     = 1122, 
    ["Soul Fire"]           = 6353,
    ["Soulburn"]            = 74434,
    --Spell damage taken debuff
    ["Curse of the Elements"]   = 1490,
    ["Earth and Moon"]          = 48506,
    ["Ebon Plague"]             = 65142,
    ["Fire Breath"]             = 34889,
    ["Lightning Breath"]        = 24844,
    ["Master Poisoner"]         = 58410,
    --Reduced casting speed debuff
    ["Curse of Tongues"]        = 1714,
    ["Lava Breath"]             = 58604,
    ["Mind-Numbing Poison"]     = 5760,
    ["Necrotic Strike"]         = 73975,
    ["Slow"]                    = 31589,
    ["Spore Cloud"]             = 50274,
    --Reduced physical damage done debuff
    ["Curse of Weakness"]       = 702,
    ["Demoralizing Roar"]       = 99,
    ["Demoralizing Screech"]    = 24423,
    ["Demoralizing Shout"]      = 1160,
    ["Scarlet Fever"]           = 81130,
    ["Vindication"]             = 26016,
}

--localize spells: use spell ids above to generate names
local spell_name
for key, value in pairs(clcInfo_Warlock.spells) do
	spell_name = GetSpellInfo(value)
	if not spell_name then spell_name = "Unknown Spell" end
	clcInfo_Warlock.spells[key] = { id = value, name = spell_name }
end

--Generic Warlock cds for all specs
--This table simplifies creating IconSpell behavior templates.
--Instead of writing an IconSpell() behavior template for each of these
--in templates.lua, we create this table which is iterated instead.
--It saves a little space and coding time.
clcInfo_Warlock.basic_cds = {
    ["Death Coil"]                      = 6789,
    ["Demon Soul"]                      = 77801,
    ["Demonic Circle Teleport"]         = 48020,
    ["Howl of Terror"]                  = 5484,
    ["Ritual of Souls"]                 = 29893,
    ["Ritual of Summoning"]             = 698,
    ["Shadow Ward"]                     = 6229,
    ["Soul Harvest"]                    = 79268,
    ["Soulburn"]                        = clcInfo_Warlock.spells["Soulburn"].id,
    ["Soulshatter"]                     = 29858,
    ["Summon Doomguard"]                = clcInfo_Warlock.spells["Summon Doomguard"].id,
    ["Summon Infernal"]                 = clcInfo_Warlock.spells["Summon Infernal"].id, 
}

-----------------FUNCTIONS---------------------------------------
--
-- Configuration functions - performance not an issue
--
-- The configuration functions are used by the module and options
-- to initialize and modify the module and settings.

-- creates the mod name. Each spec needs its own module
-- specname should be a string: the name of the spec: "Affliction"
function clcInfo_Warlock.make_modName(specname)
    return "____warlock_" .. specname
end

-- Each spec needs its own priority list of actions as well.
-- clcInfo_Warlock stores a queue for each spec.
function clcInfo_Warlock.StoreQueue(spec, q)
    clcInfo_Warlock[spec .. "_q"] = q
end

function clcInfo_Warlock.GetQueue(spec)
    return clcInfo_Warlock[spec .. "_q"]
end

-- Initializes the class module table with the name in class_module_name
-- Return the module and the module name.
function clcInfo_Warlock:init_module(specname)
    local class_module_name = self.make_modName(specname)
    local class_module = clcInfo:RegisterClassModule(class_module_name)
    return class_module, class_module_name
end

-- Update the actions queue based on the prio settings the user provides
-- This gets called when initializing the module and whenever the user
-- changes the priorities from the module options screen.
function clcInfo_Warlock.UpdateQueue(actions, tdb, spec)
	local q = {}
    local check = {} -- used to check for duplicates
    for key, value in ipairs(tdb.priority) do
        if not check[value] then
            if value ~= "none" then
                if actions[value] then
                    table.insert(q, value)
                else
                    DEFAULT_CHAT_FRAME:AddMessage("clcInfo_Warlock invalid action:" .. value)
                end
            end
        else
            DEFAULT_CHAT_FRAME:AddMessage("Ignoring duplicate action:"..key..")"..value)
        end
    end
    clcInfo_Warlock.StoreQueue(spec, q)    
end

-- register module and use default settings if version has changed
-- returns the template db, the queue, and the get and set functions for the
-- module options screen
-- called by clcInfo when templates update
-- newGET and newSET functions are defined in options.lua
function clcInfo_Warlock:OnTemplatesUpdate(class_module_name, actions, tdb_default, spec)
	local tdb = clcInfo:RegisterClassModuleTDB(class_module_name, tdb_default)
    local GetTDB, SetTDB, GetPriority, SetPriority
	if tdb then
        if not tdb.version or tdb.version < self.VERSION then
        -- fix stuff
			clcInfo.AdaptConfigAndClean(class_module_name, tdb, tdb_default)
			tdb.version = self.VERSION
		end
		
		self.UpdateQueue(actions, tdb, spec)
        GetTDB = clcInfo_Warlock.newGetTDB(tdb)
        SetTDB = clcInfo_Warlock.newSetTDB(tdb)
        GetPriority = clcInfo_Warlock.newGetPriority(tdb)
        SetPriority = clcInfo_Warlock.newSetPriority(tdb, actions, spec)
	end

    return tdb, GetTDB, SetTDB, GetPriority, SetPriority
end

-- called by clcInfo when initializing
-- returns the character db and the get and set functions for the module options screen
function clcInfo_Warlock:OnInitialize(class_module_name, db_default)
	local db = clcInfo:RegisterClassModuleDB(class_module_name, db_default)
    local GetDB = self.newGetDB(db)
    local SetDB = self.newSetDB(db)
    return db, GetDB, SetDB
end

--parse the command line to set priorities
function clcInfo_Warlock:CmdLine(tdb, args, actions, spec)
   	-- add args to options
    local num = 0
    local MAX_PRIORITY = 0
	for key, action in pairs(actions) do
        MAX_PRIORITY = MAX_PRIORITY + 1
	end

    for i, arg in ipairs(args) do
        if actions[arg] then
            if num < MAX_PRIORITY then
                num = num + 1
                tdb.priority[num] = arg
            else
                DEFAULT_CHAT_FRAME:AddMessage("clcInfo_Warlock too many cmdLine priorities.\nMax is:"..MAX_PRIORITY)
            end
        else -- unrecognized action
            DEFAULT_CHAT_FRAME:AddMessage(arg .. " not a recognized action.  Ignored.")
		end
	end
	
    -- none on the rest
    for i = num + 1, MAX_PRIORITY do
        db.priority[i] = "none"
    end

	-- redo queue
	self.UpdateQueue(actions, tdb, spec)
	clcInfo:UpdateOptions()
end

---FUNCTIONS: Frequently called---------------------------------------------------
---Below are runtime functions that are called often by the addon.  These functions
---are mostly for determining spell priorities, so they must be fast!

--make local copies of global functions to speed up their calls
local GetTime = GetTime
local GetSpellCooldown = GetSpellCooldown
local UnitAura = UnitAura
local GetQueue = clcInfo_Warlock.GetQueue
local clcInfo_debug = clcInfo.debug
local emod = clcInfo.env
local IconSpell = clcInfo.env.IconSpell
local gcdSpellId = clcInfo_Warlock.spells["Corruption"].id
local EventFrame = clcInfo_Warlock.EventFrame

-- locals for keeping track of time
local current_time, cast_left
local gcd_duration = 0

-- distance
local distance

-- mana
local mana

-- casting (really channelling) name, current ability id, next ability id
local casting_name, cur_id, next_id

--Travel time.
--Return results in seconds.  Supply distance in yards
--ANDY -- add any spells you need for tracking in flight times here.
--DOOG -- add any spells you need for tracking in flight times here.
clcInfo_Warlock.spells["Conflagrate"].GuessTravelTime = function ()
    return 0.9 -- this ranges from 0.7 to 0.9 (at least for the Backdraft buff)
end

clcInfo_Warlock.spells["Immolate"].GuessTravelTime = function ()
    return 0.55 -- this ranges from .15 to .45 irregardless of distance
end

clcInfo_Warlock.spells["Soul Fire"].GuessTravelTime = function (distance)
    return 0.25 * distance / 5 -- this varies, so guess high
end

--variables to remember last cast
local last_conflagrate_cast = 0
local last_immolate_cast = 0
local last_soul_fire_cast = 0
local last_soul_fire_distance = 40
local guess_travel_time_conflagrate = clcInfo_Warlock.spells["Conflagrate"].GuessTravelTime
local guess_travel_time_immolate = clcInfo_Warlock.spells["Immolate"].GuessTravelTime
local guess_travel_time_soul_fire = clcInfo_Warlock.spells["Soul Fire"].GuessTravelTime

--local copies to speed things up
local conflagrate = clcInfo_Warlock.spells["Conflagrate"].name
local immolate = clcInfo_Warlock.spells["Immolate"].name
local soul_fire = clcInfo_Warlock.spells["Soul Fire"].name
local IN_FLIGHT_SLOP = 0.5 -- to account for delay before tick registers on target

--in air functions. Determines if a spell is in the air
clcInfo_Warlock.spells["Conflagrate"].IsInFlight = function ()
    return GetTime() < last_conflagrate_cast + guess_travel_time_conflagrate() + IN_FLIGHT_SLOP
end

clcInfo_Warlock.spells["Immolate"].IsInFlight = function ()
    return GetTime() < last_immolate_cast + guess_travel_time_immolate() + IN_FLIGHT_SLOP
end

clcInfo_Warlock.spells["Soul Fire"].IsInFlight = function ()
    return GetTime() < 
           last_soul_fire_cast + guess_travel_time_soul_fire(last_soul_fire_distance) + IN_FLIGHT_SLOP
end

-- EVENTS!!!
-- Events are used to help determine if a spell is in flight.  When a spell with 
-- a travel time succeeds, then the time and distance to the target can be
-- recorded and the in flight time estimated.  Events are used to set the start
-- time of the spell flight.
local function handle_IN_FLIGHT_EVENTS(self, event, unit, spell)
    if unit == "player" then
        if event == "UNIT_SPELLCAST_SUCCEEDED" then
            if spell == conflagrate then
                last_conflagrate_cast = GetTime()
            elseif spell == immolate then
                last_immolate_cast = GetTime()
            elseif spell == soul_fire then
                last_soul_fire_cast = GetTime()
                _, last_soul_fire_distance = range_check:GetRange("target")
            end
        elseif event == "PLAYER_TARGET_CHANGED" then
            --if the target changes, reset immolate since it is a target DoT.
            --The new target will not have an in-flight immolate from the caster
            --until an immolate UNIT_SPELLCAST_SUCCEEDED is dropped.
            last_immolate_cast = 0
            
            --for desto locks, since Improved Soul Fire is a player buff, we
            --don't care if there's a target switch so we don't reset.
        end
    end
end

local InFlightEventFrame = CreateFrame("Frame", nil, UIParent)
InFlightEventFrame:SetScript("OnEvent", handle_IN_FLIGHT_EVENTS)
InFlightEventFrame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
InFlightEventFrame:RegisterEvent("PLAYER_TARGET_CHANGED")

------------------------------------------------------------------
-- Priority Icons
------------------------------------------------------------------

-- gets the initial status: time situation
-- resets the cur_id
-- sets current_time, casting_name, cast_left, distance, and mana.
-- Note that casting_name is nil if no spell is being channelled.
-- cast_left will either be the cast time left on casting_name or
-- the time left in the current gcd.
local function SetStatus()
    cur_id = nil

	-- current time
	current_time = GetTime()
    
    local start, gcd = GetSpellCooldown(gcdSpellId)
    if gcd > 0 then
        gcd_duration = gcd
    end

    local end_time
	casting_name, _, _, _, _, end_time = UnitCastingInfo("player")
	if casting_name then -- channelling a spell	
		cast_left = end_time/1000 - current_time
    else --gcd spell
    	cast_left = start + gcd - current_time
        if cast_left < 0 then cast_left = 0 end
    end
    
    _, distance = range_check:GetRange("target")
    if not distance then distance = 40 end

    mana = UnitPower("player", SPELL_POWER_MANA)
end

-- given a buff or debuff, the unit, and a filter, returns the time left based
-- on the current time and time left on the gcd
-- returns 0 if no such buff or debuff is present
-- actions must use this call instead of UnitAura to determine buff and debuff
-- times
function clcInfo_Warlock.GetTimeLeft(unit, spell, filter)
    local time_left = 0
    local name, expires
    name, _, _, _, _, _, expires = UnitAura(unit, spell, nil, filter)
    if name then
        time_left = expires - current_time - cast_left
        if time_left < 0 then time_left = 0 end
    end
    return time_left
end

-- given a spell id, returns the cd left on it based on the current time variables
-- actions must use this call instead of GetSpellCooldown to determine spell cds
function clcInfo_Warlock.GetCooldown(id)
	local start, duration = GetSpellCooldown(id)
	local cd = start + duration - current_time - cast_left
	if cd < 0 then return 0 end
	return cd
end

-- Calculates the target's health as a percentage.
local UnitHealth = UnitHealth
local UnitHealthMax = UnitHealthMax
function clcInfo_Warlock.CalculateTargetHealthPct()
    return floor(UnitHealth("target") / UnitHealthMax("target") * 100)
end

-- Gets time remaining on demon soul buff or 0 if no buff up
-- Wrote this in case somebody needs it.  If not we can delete it. -Deronda
local ds_felguard   = clcInfo_Warlock.spells["Demon Soul:Felguard"].name
local ds_felhunter  = clcInfo_Warlock.spells["Demon Soul:Felhunter"].name
local ds_imp        = clcInfo_Warlock.spells["Demon Soul:Imp"].name
local ds_succubus   = clcInfo_Warlock.spells["Demon Soul:Succubus"].name
local ds_voidwalker = clcInfo_Warlock.spells["Demon Soul:Voidwalker"].name
local GetTimeLeft = clcInfo_Warlock.GetTimeLeft

function clcInfo_Warlock.GetTimeLeftDemonSoul()
    local time_left
    time_left = GetTimeLeft("player", ds_felhunter, "HELPFUL")
    if time_left > 0 then return time_left end
    time_left = GetTimeLeft("player", ds_imp, "HELPFUL")
    if time_left > 0 then return time_left end
    time_left = GetTimeLeft("player", ds_felguard, "HELPFUL")
    if time_left > 0 then return time_left end
    time_left = GetTimeLeft("player", ds_succubus, "HELPFUL")
    if time_left > 0 then return time_left end
    time_left = GetTimeLeft("player", ds_voidwalker, "HELPFUL")
    return time_left
end

-- IconShadowflame
-- Maighurl's brilliant Shadowflame icon
local ShadowflameId = clcInfo_Warlock.spells["Shadowflame"].id
function emod.IconShadowflame()
    if UnitExists("target") then
        local range = range_check:GetRange("target")
        -- GetRange returns min_range and max_range.  These ranges
        -- are from a distinct set of pairs, one of which is 5, 8
        -- Since Shadowflame has a range of 8, we must check for
        -- the min_range = 5, max_range = 8 pair.
        if range <= 5 then
            return emod.IconSpell(ShadowflameId, true)
        end
    else
        return nil
    end
end

-- PRIORITIES
-- figures out which ability should be used next
-- using the queue and the module actions
-- To use this:
-- 1) call SetStatus()
-- 2) call GetNextAction() to get the first spell id to suggest.
-- 3) increment the current_time based on the cast time of the spell id suggested
-- 4) call GetNextAction() again to get the second spell id to suggest.
local function GetNextAction(actions, q)
	local n = #q
    local action_index

	-- parse once, get cooldowns, return first item in queue with 0 cd
	for i = 1, n do
        action_index = q[i]
		local action = actions[action_index]
		local cd, multiplier = action.GetCD(casting_name, cur_id, mana, distance, gcd_duration)
		if clcInfo_debug.enabled then
			clcInfo_debug:AddBoth(action_index, cd)
		end
		if cd == 0 then
			return action.id, action_index, multiplier
		end
		action.cd = cd
        action.multiplier = multiplier
	end

    -- if nothing has a cd of zero, then find the action with the
    -- lowest cooldown.  Parse again, return min cooldown
	local minQ = 1
	local minCd = actions[q[1]].cd
	for i = 2, n do
		local action_cd = actions[q[i]].cd
		if minCd > action_cd then
			minCd = action_cd
			minQ = i
		end
	end
    action_index = q[minQ]
	return actions[action_index].id, action_index, actions[action_index].multiplier
end

-- This is the driving function to determine the current and next
-- spell suggestions.
-- determines the current and next abilities
local function WarlockPriorities(actions, q)
    -- Initialize state variables.
	SetStatus()

    -- Determine current action to suggest
    local action, multiplier
	cur_id, action, multiplier = GetNextAction(actions, q)
    
    -- Adjust state variables based on the action suggestion
	local _, _, _, cost, _, _, casting_time = GetSpellInfo(cur_id)
    if casting_time == 0 then
        casting_time = gcd_duration
    else
        casting_time = casting_time/1000
    end
    mana = mana - cost --maybe add regen factor here
    current_time = current_time + cast_left + casting_time*multiplier

    -- Get next action using new state
    next_id, action = GetNextAction(actions, q)
end

local nextSkill -- next skill icon
local devName = clcInfo_Warlock.spells["Immolate"].name -- to determine range

-- Get icon for current ability
function clcInfo_Warlock.IconCurrent(actions, spec, rangePerSkill)
    local q = GetQueue(spec)
	WarlockPriorities(actions, q)
	if nextSkill then nextSkill:DoUpdate() end
  	return IconSpell(cur_id, rangePerSkill or devName)
end

-- Determine Next ability
local function NextExec()
    return IconSpell(next_id, rangePerSkill or devName)
end

-- reset next ability
local function ExecCleanupNext()
	nextSkill = nil
end

-- Get Next ability
function clcInfo_Warlock.IconNext()
	nextSkill = emod.___e
	nextSkill:SetScript("OnUpdate", nil)
	nextSkill.exec = NextExec
	nextSkill.ExecCleanup = ExecCleanupNext
end